@@ -3,6 +3,8 @@
  *  Copyright (C) 2003 Benjamin C Meyer (ben+kdelibs at meyerhome dot net)
  *  Copyright (C) 2003 Waldo Bastian <bastian@kde.org>
  *  Copyright (C) 2017 Friedrich W. H. Kossebau <kossebau@kde.org>
+ *  Copyright (C) 2020 Kevin Ottens <kevin.ottens@enioka.com>
+ *  Copyright (C) 2020 Cyril Rossi <cyril.rossi@enioka.com>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -21,6 +23,7 @@
  */
 
 #include "kconfigdialogmanager.h"
+#include "kconfigdialogmanager_p.h"
 #include "kconfigwidgets_debug.h"
 
 #include <QComboBox>
@@ -30,41 +33,14 @@
 #include <QMetaProperty>
 #include <QTimer>
 #include <QRadioButton>
-#include <QDebug>
 #include <QLayout>
 
-#include <kconfigskeleton.h>
+#include <KConfigSkeleton>
 
 typedef QHash<QString, QByteArray> MyHash;
 Q_GLOBAL_STATIC(MyHash, s_propertyMap)
 Q_GLOBAL_STATIC(MyHash, s_changedMap)
 
-class KConfigDialogManagerPrivate
-{
-
-public:
-    KConfigDialogManagerPrivate(KConfigDialogManager *q) : q(q), insideGroupBox(false) { }
-
-public:
-    KConfigDialogManager *q;
-
-    /**
-    * KConfigSkeleton object used to store settings
-     */
-    KCoreConfigSkeleton *m_conf = nullptr;
-
-    /**
-    * Dialog being managed
-     */
-    QWidget *m_dialog = nullptr;
-
-    QHash<QString, QWidget *> knownWidget;
-    QHash<QString, QWidget *> buddyWidget;
-    QSet<QWidget *> allExclusiveGroupBoxes;
-    bool insideGroupBox : 1;
-    bool trackChanges : 1;
-};
-
 KConfigDialogManager::KConfigDialogManager(QWidget *parent, KCoreConfigSkeleton *conf)
     : QObject(parent), d(new KConfigDialogManagerPrivate(this))
 {
@@ -93,6 +69,7 @@ void KConfigDialogManager::initMaps()
         s_propertyMap()->insert(QStringLiteral("KButtonGroup"), "current");
         s_propertyMap()->insert(QStringLiteral("KColorButton"), "color");
         s_propertyMap()->insert(QStringLiteral("KColorCombo"), "color");
+        s_propertyMap()->insert(QStringLiteral("KKeySequenceWidget"), "keySequence");
     }
 
     if (s_changedMap()->isEmpty()) {
@@ -143,11 +120,13 @@ QHash<QString, QByteArray> *KConfigDialogManager::propertyMap()
     return s_propertyMap();
 }
 
+#if KCONFIGWIDGETS_BUILD_DEPRECATED_SINCE(5, 32)
 QHash<QString, QByteArray> *KConfigDialogManager::changedMap()
 {
     initMaps();
     return s_changedMap();
 }
+#endif
 
 void KConfigDialogManager::init(bool trackChanges)
 {
@@ -167,7 +146,7 @@ void KConfigDialogManager::setupWidget(QWidget *widget, KConfigSkeletonItem *ite
 {
     QVariant minValue = item->minValue();
     if (minValue.isValid()) {
-        // Only q3datetimeedit is using this property we can remove it if we stop supporting Qt3Support
+        // KSelector is using this property
         if (widget->metaObject()->indexOfProperty("minValue") != -1) {
             widget->setProperty("minValue", minValue);
         }
@@ -177,7 +156,7 @@ void KConfigDialogManager::setupWidget(QWidget *widget, KConfigSkeletonItem *ite
     }
     QVariant maxValue = item->maxValue();
     if (maxValue.isValid()) {
-        // Only q3datetimeedit is using that property we can remove it if we stop supporting Qt3Support
+        // KSelector is using this property
         if (widget->metaObject()->indexOfProperty("maxValue") != -1) {
             widget->setProperty("maxValue", maxValue);
         }
@@ -212,7 +191,7 @@ void KConfigDialogManager::setupWidget(QWidget *widget, KConfigSkeletonItem *ite
             children.removeAll(gb->layout());
             const QList<QAbstractButton *> buttons = gb->findChildren<QAbstractButton *>();
             bool allAutoExclusiveDirectChildren = true;
-            foreach(QAbstractButton *button, buttons) {
+            for (QAbstractButton *button : buttons) {
                 allAutoExclusiveDirectChildren = allAutoExclusiveDirectChildren && button->autoExclusive() && button->parent() == gb;
             }
             if (allAutoExclusiveDirectChildren) {
@@ -224,20 +203,22 @@ void KConfigDialogManager::setupWidget(QWidget *widget, KConfigSkeletonItem *ite
     if (!item->isEqual(property(widget))) {
         setProperty(widget, item->property());
     }
+
+    d->updateWidgetIndicator(item->name(), widget);
 }
 
 bool KConfigDialogManager::parseChildren(const QWidget *widget, bool trackChanges)
 {
     bool valueChanged = false;
     const QList<QObject *> listOfChildren = widget->children();
-    if (listOfChildren.count() == 0) { //?? XXX
+    if (listOfChildren.isEmpty()) { //?? XXX
         return valueChanged;
     }
 
-    const QMetaMethod widgetModifiedSignal = metaObject()->method(metaObject()->indexOfSignal("widgetModified()"));
-    Q_ASSERT(widgetModifiedSignal.isValid() && metaObject()->indexOfSignal("widgetModified()")>=0);
+    const QMetaMethod onWidgetModifiedSlot = metaObject()->method(metaObject()->indexOfSlot("onWidgetModified()"));
+    Q_ASSERT(onWidgetModifiedSlot.isValid() && metaObject()->indexOfSlot("onWidgetModified()")>=0);
 
-    foreach (QObject *object, listOfChildren) {
+    for (QObject *object : listOfChildren) {
         if (!object->isWidgetType()) {
             continue;    // Skip non-widgets
         }
@@ -262,8 +243,9 @@ bool KConfigDialogManager::parseChildren(const QWidget *widget, bool trackChange
 
                     if (d->allExclusiveGroupBoxes.contains(childWidget)) {
                         const QList<QAbstractButton *> buttons = childWidget->findChildren<QAbstractButton *>();
-                        foreach(QAbstractButton *button, buttons) {
-                            connect(button, SIGNAL(toggled(bool)), this, SIGNAL(widgetModified()));
+                        for (QAbstractButton *button : buttons) {
+                            connect(button, &QAbstractButton::toggled,
+                                    this, [this] { d->onWidgetModified(); });
                         }
                     }
 
@@ -285,7 +267,7 @@ bool KConfigDialogManager::parseChildren(const QWidget *widget, bool trackChange
                                 const QMetaProperty property = metaObject->property(indexOfProperty);
                                 const QMetaMethod notifySignal = property.notifySignal();
                                 if (notifySignal.isValid()) {
-                                    connect(childWidget, notifySignal, this, widgetModifiedSignal);
+                                    connect(childWidget, notifySignal, this, onWidgetModifiedSlot);
                                     changeSignalFound = true;
                                 }
                             }
@@ -294,15 +276,15 @@ bool KConfigDialogManager::parseChildren(const QWidget *widget, bool trackChange
                         }
                     } else {
                         connect(childWidget, propertyChangeSignal,
-                                this, SIGNAL(widgetModified()));
+                                this, SLOT(onWidgetModified()));
                         changeSignalFound = true;
                     }
 
                     if (changeSignalFound) {
                         QComboBox *cb = qobject_cast<QComboBox *>(childWidget);
                         if (cb && cb->isEditable())
-                            connect(cb, SIGNAL(editTextChanged(QString)),
-                                    this, SIGNAL(widgetModified()));
+                            connect(cb, &QComboBox::editTextChanged,
+                                    this, &KConfigDialogManager::widgetModified);
                     }
                 }
                 QGroupBox *gb = qobject_cast<QGroupBox *>(childWidget);
@@ -384,7 +366,8 @@ void KConfigDialogManager::updateWidgets()
     blockSignals(bSignalsBlocked);
 
     if (changed) {
-        QTimer::singleShot(0, this, SIGNAL(widgetModified()));
+        QTimer::singleShot(0, this, &KConfigDialogManager::widgetModified);
+        d->updateAllWidgetIndicators();
     }
 }
 
@@ -393,6 +376,12 @@ void KConfigDialogManager::updateWidgetsDefault()
     bool bUseDefaults = d->m_conf->useDefaults(true);
     updateWidgets();
     d->m_conf->useDefaults(bUseDefaults);
+    d->updateAllWidgetIndicators();
+}
+
+void KConfigDialogManager::setDefaultsIndicatorsVisible(bool enabled)
+{
+    d->setDefaultsIndicatorsVisible(enabled);
 }
 
 void KConfigDialogManager::updateSettings()
@@ -420,6 +409,7 @@ void KConfigDialogManager::updateSettings()
     if (changed) {
         d->m_conf->save();
         emit settingsChanged();
+        d->updateAllWidgetIndicators();
     }
 }
 
@@ -598,3 +588,74 @@ bool KConfigDialogManager::isDefault() const
     return result;
 }
 
+KConfigDialogManagerPrivate::KConfigDialogManagerPrivate(KConfigDialogManager *q)
+    : q(q)
+    , insideGroupBox(false)
+    , defaultsIndicatorsVisible(false)
+{
+}
+
+void KConfigDialogManagerPrivate::setDefaultsIndicatorsVisible(bool enabled)
+{
+    if (defaultsIndicatorsVisible != enabled) {
+        defaultsIndicatorsVisible = enabled;
+        updateAllWidgetIndicators();
+    }
+}
+
+void KConfigDialogManagerPrivate::onWidgetModified()
+{
+    const auto widget = qobject_cast<QWidget*>(q->sender());
+    Q_ASSERT(widget);
+
+    if (!widget->objectName().startsWith("kcfg_")) {
+        Q_ASSERT(widget->parent()->objectName().startsWith("kcfg_"));
+        const auto configId = widget->parent()->objectName().mid(5);
+        const auto parent = qobject_cast<QWidget*>(widget->parent());
+        Q_ASSERT(parent);
+        updateWidgetIndicator(configId, parent);
+    } else {
+        const auto configId = widget->objectName().mid(5);
+        updateWidgetIndicator(configId, widget);
+    }
+    emit q->widgetModified();
+}
+
+void KConfigDialogManagerPrivate::updateWidgetIndicator(const QString &configId, QWidget *widget)
+{
+    const auto item = m_conf->findItem(configId);
+    Q_ASSERT(item);
+
+    const auto widgetValue = q->property(widget);
+    const auto defaultValue = [item] {
+        item->swapDefault();
+        const auto value = item->property();
+        item->swapDefault();
+        return value;
+    }();
+
+    const auto defaulted = widgetValue == defaultValue;
+
+    if (allExclusiveGroupBoxes.contains(widget)) {
+        const QList<QAbstractButton *> buttons = widget->findChildren<QAbstractButton *>();
+        for (int i = 0; i < buttons.count(); i++) {
+            const auto value = widgetValue.toInt() == i && !defaulted && defaultsIndicatorsVisible;
+            buttons.at(i)->setProperty("_kde_highlight_neutral", value);
+            buttons.at(i)->update();
+        }
+    } else {
+        widget->setProperty("_kde_highlight_neutral", !defaulted && defaultsIndicatorsVisible);
+        widget->update();
+    }
+}
+
+void KConfigDialogManagerPrivate::updateAllWidgetIndicators()
+{
+    QHashIterator<QString, QWidget *> it(knownWidget);
+    while (it.hasNext()) {
+        it.next();
+        updateWidgetIndicator(it.key(), it.value());
+    }
+}
+
+#include "moc_kconfigdialogmanager.cpp"
